@@ -50,6 +50,8 @@
  */
 typedef struct {
     GQueue *queue;
+    GQueue *running_queue;
+    
     //GList *cursor;
     guint idle_id;
     guint gconf_notify_id;
@@ -72,7 +74,8 @@ typedef struct {
     OssoThumbnailFlags flags;
     OssoThumbnailFactoryFinishedCallback callback;
     gpointer user_data;
-
+    gboolean canceled;
+    
     // Data for spawned process (when necessary)
     guint thumb_width, thumb_height;
     gchar *temp_file;
@@ -120,6 +123,8 @@ static GQuark app_quark = 0;
 
 static gboolean show_debug = FALSE;
 
+#define NR_MP3_OPTIONS 5
+
 // Private functions
 
 // Utility functions
@@ -252,7 +257,9 @@ static gboolean on_queue_change()
 
 static void run_callback(ThumbsItem *item, GdkPixbuf *pixbuf, GError *error)
 {
-    item->callback(THUMBS_HANDLE(item), item->user_data, pixbuf, error);
+    if(!item->canceled) {
+        item->callback(THUMBS_HANDLE(item), item->user_data, pixbuf, error);
+    }
 }
 
 static GdkPixbuf* load_thumb_file(gchar *file, time_t *mtime, const gchar **uri,
@@ -478,6 +485,8 @@ static gint process_thumb(ThumbsItem *item, GError **error)
 
     // Go, horsey, go!
     factory->nprocesses++;
+    
+    g_queue_push_tail(factory->running_queue, item);
 
     on_queue_change();
 
@@ -515,6 +524,8 @@ static void watch_func(GPid pid, gint status, gpointer data)
 {
     ThumbsItem *item = (ThumbsItem*)data;
 
+    g_queue_remove(factory->running_queue, item);
+    
     factory->nprocesses--;
 
     g_spawn_close_pid(pid);
@@ -528,7 +539,9 @@ static void watch_func(GPid pid, gint status, gpointer data)
             if(show_debug)
                 g_message("Thumbnail hot off the press: %s", item->thumb_file);
 
-            load_final_thumb(item, NULL);
+            if(!item->canceled) {
+                load_final_thumb(item, NULL);
+            }
         } else {
             // Houston, we have a problem
             if(rename(item->temp_file, item->fail_file))
@@ -794,7 +807,8 @@ static gboolean thumbs_init()
 
         factory = g_new(ThumbsFactory, 1);
         factory->queue = g_queue_new();
-        //factory->cursor = NULL;
+        factory->running_queue = g_queue_new();
+                
         factory->idle_id = 0;
         factory->gconf_notify_id = 0;
 
@@ -813,13 +827,59 @@ static void update_thumb(gchar *src_file, gchar *dest_file,
                          const gchar *uri, time_t mtime)
 {
     GdkPixbuf *pixbuf;
+    gchar *mp3;
+    gchar **keys, **values;
 
     pixbuf = load_thumb_file(src_file, NULL, NULL, NULL);
     if(!pixbuf) {
         return;
     }
 
-    save_thumb_file(pixbuf, dest_file, mtime, uri);
+    mp3 = gdk_pixbuf_get_option(pixbuf, MP3_OPTION);
+
+    if (mp3)
+    {
+        /* keep MP3 options when saving */
+        gchar *noimage, *title, *artist, *album;
+        gint i = 0;
+        keys = g_new0(char *, NR_MP3_OPTIONS + 1);
+        keys[i] = g_strdup (MP3_OPTION);
+        values = g_new0(char *, NR_MP3_OPTIONS + 1);
+        values[i] = g_strdup (mp3);
+        i++;
+        
+        if ( noimage = gdk_pixbuf_get_option(pixbuf, MP3_NOIMAGE_OPTION) )
+        {
+            keys[i] = g_strdup (MP3_NOIMAGE_OPTION);
+            values[i] = g_strdup (noimage);
+            i++;
+        }
+        if ( title = gdk_pixbuf_get_option(pixbuf, MP3_TITLE_OPTION) )
+        {
+            keys[i] = g_strdup (MP3_TITLE_OPTION);
+            values[i] = g_strdup (title);
+            i++;
+        }
+        if ( artist = gdk_pixbuf_get_option(pixbuf, MP3_ARTIST_OPTION) )
+        {
+            keys[i] = g_strdup (MP3_ARTIST_OPTION);
+            values[i] = g_strdup (artist);
+            i++;
+        }   
+        if ( album = gdk_pixbuf_get_option(pixbuf, MP3_ALBUM_OPTION) )
+        {
+            keys[i] = g_strdup (MP3_ALBUM_OPTION);
+            values[i] = g_strdup (album);
+            i++;
+        }
+        
+        save_thumb_file_meta(pixbuf, dest_file, mtime, uri, keys, values);
+
+        g_strfreev(keys);
+        g_strfreev(values);
+    }
+    else
+        save_thumb_file(pixbuf, dest_file, mtime, uri);
 
     gdk_pixbuf_unref(pixbuf);
 }
@@ -1060,6 +1120,7 @@ OssoThumbnailFactoryHandle osso_thumbnail_factory_load_custom(
     item->callback = callback;
     item->user_data = user_data;
     item->flags = flags;
+    item->canceled = FALSE;
 
     item->temp_file = item->thumb_file = item->fail_file = NULL;
     item->thumb_width = item->thumb_height = 0;
@@ -1100,6 +1161,17 @@ void osso_thumbnail_factory_cancel(OssoThumbnailFactoryHandle handle)
         thumb_item_free(item);
 
         on_queue_change();
+    } else {    
+        // Not found in work queue,  maybe it is running?
+        lst = g_queue_find(factory->running_queue, item);
+        
+        // Yes, the handle is valid
+        if(lst) {
+            item->canceled = TRUE;
+            // Item will be freed by process notify func
+        } else {
+            g_warning("Thumbnail cancel on handle that doesn't exist in queue: %08X", (unsigned int)handle);
+        }
     }
 }
 
